"""Autoscalers: perform autoscaling by monitoring metrics."""
import bisect
import collections
import dataclasses
import enum
import math
import time
import typing
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Union

from sky import sky_logging
from sky.serve import constants
from sky.serve import serve_state

if typing.TYPE_CHECKING:
    from sky.serve import replica_managers
    from sky.serve import service_spec

logger = sky_logging.init_logger(__name__)


class AutoscalerDecisionOperator(enum.Enum):
    SCALE_UP = 'scale_up'
    SCALE_DOWN = 'scale_down'


@dataclasses.dataclass
class AutoscalerDecision:
    """Autoscaling decisions.

    |------------------------------------------------------------------------|
    | Operator   | TargetType                | Meaning                       |
    |------------|---------------------------|-------------------------------|
    | SCALE_UP   | Optional[Dict[str, Any]   | Resource override to add      |
    |------------|---------------------------|-------------------------------|
    | SCALE_DOWN | int                       | Replica id to remove          |
    |------------------------------------------------------------------------|
    """
    operator: AutoscalerDecisionOperator
    target: Union[Optional[Dict[str, Any]], int]

    # TODO(MaoZiming): Add a doc to elaborate on autoscaling policies.
    def __init__(self, operator: AutoscalerDecisionOperator,
                 target: Union[Optional[Dict[str, Any]], int]):

        assert (operator == AutoscalerDecisionOperator.SCALE_UP and
                (target is None or isinstance(target, dict))) or (
                    operator == AutoscalerDecisionOperator.SCALE_DOWN and
                    isinstance(target, int))
        self.operator = operator
        self.target = target

    def __repr__(self) -> str:
        return f'AutoscalerDecision({self.operator}, {self.target})'


class Autoscaler:
    """Abstract class for autoscalers."""

    def __init__(self, service_name: str,
                 spec: 'service_spec.SkyServiceSpec') -> None:
        """Initialize the autoscaler.

        Variables:
            min_replicas: Minimum number of replicas.
            max_replicas: Maximum number of replicas. Default to fixed
                number of replicas, i.e. min_replicas == max_replicas.
            target_num_replicas: Target number of replicas output by autoscaler.
            latest_version: latest version of the service.
        """
        self._service_name: str = service_name
        self.min_replicas: int = spec.min_replicas
        self.max_replicas: int = (spec.max_replicas if spec.max_replicas
                                  is not None else spec.min_replicas)
        # Target number of replicas is initialized to min replicas
        self.target_num_replicas: int = spec.min_replicas
        self.latest_version: int = constants.INITIAL_VERSION

    def update_version(self, version: int,
                       spec: 'service_spec.SkyServiceSpec') -> None:
        if version <= self.latest_version:
            logger.error(f'Invalid version: {version}, '
                         f'latest version: {self.latest_version}')
            return
        self.latest_version = version
        self.min_replicas = spec.min_replicas
        self.max_replicas = (spec.max_replicas if spec.max_replicas is not None
                             else spec.min_replicas)
        # Reclip self.target_num_replicas with new min and max replicas.
        self.target_num_replicas = max(
            self.min_replicas, min(self.max_replicas, self.target_num_replicas))

    def collect_request_information(
            self, request_aggregator_info: Dict[str, Any]) -> None:
        """Collect request information from aggregator for autoscaling."""
        raise NotImplementedError

    def evaluate_scaling(
        self,
        replica_infos: List['replica_managers.ReplicaInfo'],
    ) -> List[AutoscalerDecision]:
        """Evaluate autoscale options based on replica information."""
        raise NotImplementedError

    @classmethod
    def from_spec(cls, service_name: str,
                  spec: 'service_spec.SkyServiceSpec') -> 'Autoscaler':
        # TODO(MaoZiming): use NAME to get the class.
        if spec.use_ondemand_fallback:
            return FallbackRequestRateAutoscaler(service_name, spec)
        else:
            return RequestRateAutoscaler(service_name, spec)

    def get_latest_version_with_min_replicas(
        self, replica_infos: Iterable['replica_managers.ReplicaInfo']
    ) -> Optional[int]:
        """Get the latest version with at least min_replicas replicas."""
        raise NotImplementedError

    def dump_dynamic_states(self) -> Dict[str, Any]:
        """Dump dynamic states from autoscaler."""
        raise NotImplementedError

    def load_dynamic_states(self, dynamic_states: Dict[str, Any]) -> None:
        """Load dynamic states to autoscaler."""
        raise NotImplementedError


class RequestRateAutoscaler(Autoscaler):
    """RequestRateAutoscaler: Autoscale according to request rate.

    Scales when the number of requests per replica in the given interval
    is above or below the target qps per replica. The instance can be
    either spot or on-demand, but not both.
    """

    def __init__(self, service_name: str,
                 spec: 'service_spec.SkyServiceSpec') -> None:
        """Initialize the request rate autoscaler.

        Variables:
            target_qps_per_replica: Target qps per replica for autoscaling.
            qps_window_size: Window size for qps calculating.
            request_timestamps: All request timestamps within the window.
            upscale_counter: counter for upscale number of replicas.
            downscale_counter: counter for downscale number of replicas.
            scale_up_consecutive_periods: period for scaling up.
            scale_down_consecutive_periods: period for scaling down.
        """
        super().__init__(service_name, spec)
        self.target_qps_per_replica: Optional[
            float] = spec.target_qps_per_replica
        self.qps_window_size: int = constants.AUTOSCALER_QPS_WINDOW_SIZE_SECONDS
        self.request_timestamps: List[float] = []
        self.upscale_counter: int = 0
        self.downscale_counter: int = 0
        upscale_delay_seconds = (
            spec.upscale_delay_seconds if spec.upscale_delay_seconds is not None
            else constants.AUTOSCALER_DEFAULT_UPSCALE_DELAY_SECONDS)
        self.scale_up_consecutive_periods: int = int(
            upscale_delay_seconds /
            constants.AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS)
        downscale_delay_seconds = (
            spec.downscale_delay_seconds
            if spec.downscale_delay_seconds is not None else
            constants.AUTOSCALER_DEFAULT_DOWNSCALE_DELAY_SECONDS)
        self.scale_down_consecutive_periods: int = int(
            downscale_delay_seconds /
            constants.AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS)

    def _cal_target_num_replicas_based_on_qps(self) -> int:
        # Recalculate target_num_replicas based on QPS.
        # Reclip self.target_num_replicas with new min and max replicas.
        if self.target_qps_per_replica is None:
            return self.min_replicas
        target_num_replicas = math.ceil(
            len(self.request_timestamps) / self.qps_window_size /
            self.target_qps_per_replica)
        return max(self.min_replicas, min(self.max_replicas,
                                          target_num_replicas))

    def update_version(self, version: int,
                       spec: 'service_spec.SkyServiceSpec') -> None:
        super().update_version(version, spec)
        self.target_qps_per_replica = spec.target_qps_per_replica
        upscale_delay_seconds = (
            spec.upscale_delay_seconds if spec.upscale_delay_seconds is not None
            else constants.AUTOSCALER_DEFAULT_UPSCALE_DELAY_SECONDS)
        self.scale_up_consecutive_periods = int(
            upscale_delay_seconds /
            constants.AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS)
        downscale_delay_seconds = (
            spec.downscale_delay_seconds
            if spec.downscale_delay_seconds is not None else
            constants.AUTOSCALER_DEFAULT_DOWNSCALE_DELAY_SECONDS)
        self.scale_down_consecutive_periods = int(
            downscale_delay_seconds /
            constants.AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS)

        # We directly set the target_num_replicas here instead of
        # calling `_set_target_num_replica_with_hysteresis` to have the replicas
        # quickly scale after each update.
        self.target_num_replicas = self._cal_target_num_replicas_based_on_qps()
        # Cleanup hysteretic counters.
        self.upscale_counter = 0
        self.downscale_counter = 0

    def collect_request_information(
            self, request_aggregator_info: Dict[str, Any]) -> None:
        """Collect request information from aggregator for autoscaling.

        request_aggregator_info should be a dict with the following format:

        {
            'timestamps': [timestamp1 (float), timestamp2 (float), ...]
        }
        """
        self.request_timestamps.extend(
            request_aggregator_info.get('timestamps', []))
        current_time = time.time()
        index = bisect.bisect_left(self.request_timestamps,
                                   current_time - self.qps_window_size)
        self.request_timestamps = self.request_timestamps[index:]

    def _set_target_num_replica_with_hysteresis(self) -> None:
        """Set target_num_replicas based on request rate with hysteresis."""
        # Keep self.target_num_replicas unchange when autoscaling
        # is not enabled, i.e. self.target_qps_per_replica is None.
        # In this case, self.target_num_replicas will be min_replicas.
        if self.target_qps_per_replica is None:
            return

        # Convert to requests per second.
        target_num_replicas = self._cal_target_num_replicas_based_on_qps()
        old_target_num_replicas = self.target_num_replicas

        # Faster scale up when there is no replica.
        if self.target_num_replicas == 0:
            self.target_num_replicas = target_num_replicas
        elif target_num_replicas > self.target_num_replicas:
            self.upscale_counter += 1
            self.downscale_counter = 0
            if self.upscale_counter >= self.scale_up_consecutive_periods:
                self.upscale_counter = 0
                self.target_num_replicas = target_num_replicas
        elif target_num_replicas < self.target_num_replicas:
            self.downscale_counter += 1
            self.upscale_counter = 0
            if self.downscale_counter >= self.scale_down_consecutive_periods:
                self.downscale_counter = 0
                self.target_num_replicas = target_num_replicas
        else:
            self.upscale_counter = self.downscale_counter = 0

        num_requests_per_second = len(
            self.request_timestamps) / self.qps_window_size
        logger.info(
            f'Requests per second: {num_requests_per_second}. '
            f'Current target number of replicas: {old_target_num_replicas}. '
            f'Final target number of replicas: {self.target_num_replicas}. '
            f'Upscale counter: {self.upscale_counter}/'
            f'{self.scale_up_consecutive_periods}. '
            f'Downscale counter: {self.downscale_counter}/'
            f'{self.scale_down_consecutive_periods}')

    @classmethod
    def _select_replicas_to_scale_down(
            cls, num_limit: int,
            replica_infos: Iterable['replica_managers.ReplicaInfo']
    ) -> List[int]:

        status_order = serve_state.ReplicaStatus.scale_down_decision_order()
        # Also sort by provisioned time (indicated by replica_id) to make sure
        # we terminate the replicas that starts provisioning later first
        replica_infos_sorted = sorted(
            replica_infos,
            key=lambda info: (
                status_order.index(info.status)
                # Use -1 for other status that will be terminated first.
                # Including: NOT_READY, SHUTTING_DOWN, FAILED,
                # FAILED_CLEANUP, PREEMPTED, UNKNOWN.
                if info.status in status_order else -1,
                # `-info.replica_id` is to furhter sort the replicas with the
                # same state by the time it starts launching. In that case,
                # if two replicas are both in PROVISIONING state, we will scale
                # down the one that starts provisioning later, i.e., it will
                # take a longer time for it to be READY.
                -info.replica_id))

        return [info.replica_id for info in replica_infos_sorted][:num_limit]

    def get_decision_interval(self) -> int:
        # Reduce autoscaler interval when target_num_replicas = 0.
        # This will happen when min_replicas = 0 and no traffic.
        if self.target_num_replicas == 0:
            return constants.AUTOSCALER_NO_REPLICA_DECISION_INTERVAL_SECONDS
        else:
            return constants.AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS

    def get_latest_version_with_min_replicas(
        self, replica_infos: Iterable['replica_managers.ReplicaInfo']
    ) -> Optional[int]:
        # Find the latest version with at least min_replicas replicas.
        version2count: DefaultDict[int, int] = collections.defaultdict(int)
        for info in replica_infos:
            if info.is_ready:
                version2count[info.version] += 1

        version = self.latest_version
        while version >= constants.INITIAL_VERSION:
            spec = serve_state.get_spec(self._service_name, version)
            if (spec is not None and
                    version2count[version] >= spec.min_replicas):
                return version
            version -= 1
        return None

    def select_outdated_replicas_to_scale_down(
            self, replica_infos: Iterable['replica_managers.ReplicaInfo']
    ) -> List[int]:

        latest_version_with_min_replicas = (
            self.get_latest_version_with_min_replicas(replica_infos))

        # Select replicas earlier than latest_version_with_min_replicas to scale
        # down.
        all_replica_ids_to_scale_down: List[int] = []
        for info in replica_infos:
            if (latest_version_with_min_replicas is not None and
                    info.version < latest_version_with_min_replicas):
                all_replica_ids_to_scale_down.append(info.replica_id)

        return all_replica_ids_to_scale_down

    def evaluate_scaling(
        self,
        replica_infos: List['replica_managers.ReplicaInfo'],
    ) -> List[AutoscalerDecision]:
        """Evaluate Autoscaling decisions based on replica information.
        If the number of launched replicas is less than the target,
        Trigger a scale up. Else, trigger a scale down.

        For future compatibility, we return a list of AutoscalerDecision.
        Scale-up could include both spot and on-demand, each with a resource
        override dict. Active migration could require returning both SCALE_UP
        and SCALE_DOWN.
        """
        latest_provisioning_and_launched_replicas: List[
            'replica_managers.ReplicaInfo'] = []

        for info in replica_infos:
            if info.version == self.latest_version:
                if info.is_provisioning_or_launched:
                    latest_provisioning_and_launched_replicas.append(info)

        self._set_target_num_replica_with_hysteresis()

        scaling_options: List[AutoscalerDecision] = []
        all_replica_ids_to_scale_down: List[int] = []

        # Case 1. Once there is min_replicas number of
        # ready new replicas, we will direct all traffic to them,
        # we can scale down all old replicas.
        all_replica_ids_to_scale_down.extend(
            self.select_outdated_replicas_to_scale_down(replica_infos))

        # Case 2. when latest_provisioning_and_launched_replicas is less
        # than num_to_provision, we always scale up new replicas.
        if len(latest_provisioning_and_launched_replicas
              ) < self.target_num_replicas:
            num_replicas_to_scale_up = (
                self.target_num_replicas -
                len(latest_provisioning_and_launched_replicas))

            for _ in range(num_replicas_to_scale_up):
                scaling_options.append(
                    AutoscalerDecision(AutoscalerDecisionOperator.SCALE_UP,
                                       target=None))

        # Case 3: when latest_provisioning_and_launched_replicas is more
        # than self.target_num_replicas, we scale down new replicas.
        if len(latest_provisioning_and_launched_replicas
              ) > self.target_num_replicas:
            num_replicas_to_scale_down = (
                len(latest_provisioning_and_launched_replicas) -
                self.target_num_replicas)
            all_replica_ids_to_scale_down.extend(
                RequestRateAutoscaler._select_replicas_to_scale_down(
                    num_limit=num_replicas_to_scale_down,
                    replica_infos=latest_provisioning_and_launched_replicas))

        for replica_id in all_replica_ids_to_scale_down:
            scaling_options.append(
                AutoscalerDecision(AutoscalerDecisionOperator.SCALE_DOWN,
                                   target=replica_id))

        if not scaling_options:
            logger.info('No scaling needed.')
        return scaling_options

    def dump_dynamic_states(self) -> Dict[str, Any]:
        return {
            'request_timestamps': self.request_timestamps,
        }

    def load_dynamic_states(self, dynamic_states: Dict[str, Any]) -> None:
        if 'request_timestamps' in dynamic_states:
            self.request_timestamps = dynamic_states.pop('request_timestamps')
        if dynamic_states:
            logger.info(f'Remaining dynamic states: {dynamic_states}')


class FallbackRequestRateAutoscaler(RequestRateAutoscaler):
    """FallbackRequestRateAutoscaler

    Autoscale based on request rate. It adds additional ability to
    RequestRateAutoscaler for having spot with on-demand fallback.

    When spec.base_ondemand_fallback_replicas is set, we make sure
    there are at least spec.base_ondemand_fallback_replicas on-demands
    to be always there to provide basic gurantee for the availability.

    When spec.dynamic_ondemand_fallback is set, on-demand instances
    will be scheduled to provision for any preempted spot instance, i.e.,
    on-demand instance are used as dynamic fallback of spot.
    """

    def __init__(self, service_name: str,
                 spec: 'service_spec.SkyServiceSpec') -> None:
        super().__init__(service_name, spec)
        self.base_ondemand_fallback_replicas: int = (
            spec.base_ondemand_fallback_replicas
            if spec.base_ondemand_fallback_replicas is not None else 0)
        # Assert: Either dynamic_ondemand_fallback is set
        # or base_ondemand_fallback_replicas is greater than 0.
        assert spec.use_ondemand_fallback
        self.dynamic_ondemand_fallback: bool = (
            spec.dynamic_ondemand_fallback
            if spec.dynamic_ondemand_fallback is not None else False)

    def update_version(self, version: int,
                       spec: 'service_spec.SkyServiceSpec') -> None:
        super().update_version(version, spec)
        self.base_ondemand_fallback_replicas = (
            spec.base_ondemand_fallback_replicas
            if spec.base_ondemand_fallback_replicas is not None else 0)
        # Assert: Either dynamic_ondemand_fallback is set
        # or base_ondemand_fallback_replicas is greater than 0.
        assert spec.use_ondemand_fallback
        self.dynamic_ondemand_fallback = (spec.dynamic_ondemand_fallback
                                          if spec.dynamic_ondemand_fallback
                                          is not None else False)

    # spot_recovery field is checked earlier in core
    def _get_spot_resources_override_dict(self) -> Dict[str, Any]:
        return {'use_spot': True}

    def _get_ondemand_resources_override_dict(self) -> Dict[str, Any]:
        return {'use_spot': False}

    def evaluate_scaling(
        self,
        replica_infos: List['replica_managers.ReplicaInfo'],
    ) -> List[AutoscalerDecision]:

        latest_provisioning_and_launched_replicas = list(
            filter(
                lambda info: info.is_provisioning_or_launched and info.version
                == self.latest_version, replica_infos))

        self._set_target_num_replica_with_hysteresis()
        num_provisioning_and_launched_spot, num_ready_spot = 0, 0
        num_provisioning_and_launched_ondemand, num_ready_ondemand = 0, 0

        for info in latest_provisioning_and_launched_replicas:
            if info.is_spot:
                if info.status == serve_state.ReplicaStatus.READY:
                    num_ready_spot += 1
                num_provisioning_and_launched_spot += 1
            else:
                if info.status == serve_state.ReplicaStatus.READY:
                    num_ready_ondemand += 1
                num_provisioning_and_launched_ondemand += 1

        logger.info(
            'Number of alive spot instances: '
            f'{num_provisioning_and_launched_spot}, '
            f'Number of ready spot instances: {num_ready_spot}, '
            'Number of alive on-demand instances: '
            f' {num_provisioning_and_launched_ondemand}, '
            f'Number of ready on-demand instances: {num_ready_ondemand}')

        scaling_options: List[AutoscalerDecision] = []
        all_replica_ids_to_scale_down: List[int] = []

        # Once there is min_replicas number of ready new replicas, we will
        # direct all traffic to them, we can scale down all old replicas.
        all_replica_ids_to_scale_down.extend(
            self.select_outdated_replicas_to_scale_down(replica_infos))

        # Decide how many spot instances to launch.
        num_spot_to_provision = (self.target_num_replicas -
                                 self.base_ondemand_fallback_replicas)
        if num_provisioning_and_launched_spot < num_spot_to_provision:
            # Not enough spot instances, scale up.
            num_spot_to_scale_up = (num_spot_to_provision -
                                    num_provisioning_and_launched_spot)
            for _ in range(num_spot_to_scale_up):
                scaling_options.append(
                    AutoscalerDecision(
                        AutoscalerDecisionOperator.SCALE_UP,
                        target=self._get_spot_resources_override_dict()))
        elif num_provisioning_and_launched_spot > num_spot_to_provision:
            # Too many spot instances, scale down.
            # Get the replica to scale down with _select_replicas_to_scale_down
            num_spot_to_scale_down = (num_provisioning_and_launched_spot -
                                      num_spot_to_provision)
            all_replica_ids_to_scale_down.extend(
                RequestRateAutoscaler._select_replicas_to_scale_down(
                    num_spot_to_scale_down,
                    filter(lambda info: info.is_spot,
                           latest_provisioning_and_launched_replicas)))

        # Decide how many on-demand instances to launch.
        num_ondemand_to_provision = self.base_ondemand_fallback_replicas
        if self.dynamic_ondemand_fallback:
            # `num_ready_spot` instead of `num_provisioning_and_launched_spot`
            # because the provisioning spot can fail to UP due to the capacity
            # issue, and on-demand should fill the gap between the required
            # number of spot and ready spot.
            num_ondemand_to_provision += (num_spot_to_provision -
                                          num_ready_spot)

        if num_ondemand_to_provision > num_provisioning_and_launched_ondemand:
            for _ in range(num_ondemand_to_provision -
                           num_provisioning_and_launched_ondemand):
                scaling_options.append(
                    AutoscalerDecision(
                        AutoscalerDecisionOperator.SCALE_UP,
                        target=self._get_ondemand_resources_override_dict()))
        else:
            all_replica_ids_to_scale_down.extend(
                RequestRateAutoscaler._select_replicas_to_scale_down(
                    num_provisioning_and_launched_ondemand -
                    num_ondemand_to_provision,
                    filter(lambda info: not info.is_spot,
                           latest_provisioning_and_launched_replicas)))

        for replica_id in all_replica_ids_to_scale_down:
            scaling_options.append(
                AutoscalerDecision(AutoscalerDecisionOperator.SCALE_DOWN,
                                   target=replica_id))

        return scaling_options



"""The SkyPilot package."""
import os
import subprocess

# Replaced with the current commit when building the wheels.
_SKYPILOT_COMMIT_SHA = '14080a05a5686422417998b5b738900cc466e72c'


def get_git_commit():
    if 'SKYPILOT_COMMIT_SHA' not in _SKYPILOT_COMMIT_SHA:
        # This is a release build, so we don't need to get the commit hash from
        # git, as it's already been set.
        return _SKYPILOT_COMMIT_SHA

    # This is a development build (pip install -e .), so we need to get the
    # commit hash from git.
    try:
        cwd = os.path.dirname(__file__)
        commit_hash = subprocess.check_output(
            ['git', 'rev-parse', 'HEAD'],
            cwd=cwd,
            universal_newlines=True,
            stderr=subprocess.DEVNULL).strip()
        changes = subprocess.check_output(['git', 'status', '--porcelain'],
                                          cwd=cwd,
                                          universal_newlines=True,
                                          stderr=subprocess.DEVNULL).strip()
        if changes:
            commit_hash += '-dirty'
        return commit_hash
    except Exception:  # pylint: disable=broad-except
        return _SKYPILOT_COMMIT_SHA


__commit__ = get_git_commit()
__version__ = '1.0.0.dev20240305'
__root_dir__ = os.path.dirname(os.path.abspath(__file__))

# Keep this order to avoid cyclic imports
# pylint: disable=wrong-import-position
from sky import backends
from sky import benchmark
from sky import clouds
from sky.clouds.service_catalog import list_accelerators
from sky.core import autostop
from sky.core import cancel
from sky.core import cost_report
from sky.core import down
from sky.core import download_logs
from sky.core import job_status
from sky.core import queue
from sky.core import spot_cancel
from sky.core import spot_queue
from sky.core import spot_status
from sky.core import start
from sky.core import status
from sky.core import stop
from sky.core import storage_delete
from sky.core import storage_ls
from sky.core import tail_logs
from sky.dag import Dag
from sky.data import Storage
from sky.data import StorageMode
from sky.data import StoreType
from sky.execution import exec  # pylint: disable=redefined-builtin
from sky.execution import launch
from sky.execution import spot_launch
from sky.optimizer import Optimizer
from sky.optimizer import OptimizeTarget
from sky.resources import Resources
from sky.skylet.job_lib import JobStatus
from sky.status_lib import ClusterStatus
from sky.task import Task

# Aliases.
IBM = clouds.IBM
AWS = clouds.AWS
Azure = clouds.Azure
Cudo = clouds.Cudo
GCP = clouds.GCP
Lambda = clouds.Lambda
SCP = clouds.SCP
Kubernetes = clouds.Kubernetes
OCI = clouds.OCI
RunPod = clouds.RunPod
Vsphere = clouds.Vsphere
Fluidstack = clouds.Fluidstack
optimize = Optimizer.optimize

__all__ = [
    '__version__',
    'AWS',
    'Azure',
    'Cudo',
    'GCP',
    'IBM',
    'Kubernetes',
    'Lambda',
    'OCI',
    'RunPod',
    'SCP',
    'Vsphere',
    'Fluidstack',
    'Optimizer',
    'OptimizeTarget',
    'backends',
    'benchmark',
    'list_accelerators',
    '__root_dir__',
    'Storage',
    'StorageMode',
    'StoreType',
    'ClusterStatus',
    'JobStatus',
    # APIs
    'Dag',
    'Task',
    'Resources',
    # execution APIs
    'launch',
    'exec',
    'spot_launch',
    # core APIs
    'status',
    'start',
    'stop',
    'down',
    'autostop',
    'cost_report',
    # core APIs Job Management
    'queue',
    'cancel',
    'tail_logs',
    'download_logs',
    'job_status',
    # core APIs Spot Job Management
    'spot_queue',
    'spot_status',  # Deprecated (alias for spot_queue)
    'spot_cancel',
    # core APIs Storage Management
    'storage_ls',
    'storage_delete',
]



"""Clouds in Sky."""
from sky.clouds.cloud import Cloud
from sky.clouds.cloud import cloud_in_list
from sky.clouds.cloud import CloudImplementationFeatures
from sky.clouds.cloud import ProvisionerVersion
from sky.clouds.cloud import Region
from sky.clouds.cloud import StatusVersion
from sky.clouds.cloud import Zone
from sky.clouds.cloud_registry import CLOUD_REGISTRY

# NOTE: import the above first to avoid circular imports.
# isort: split
from sky.clouds.aws import AWS
from sky.clouds.azure import Azure
from sky.clouds.cudo import Cudo
from sky.clouds.fluidstack import Fluidstack
from sky.clouds.gcp import GCP
from sky.clouds.ibm import IBM
from sky.clouds.kubernetes import Kubernetes
from sky.clouds.lambda_cloud import Lambda
from sky.clouds.oci import OCI
from sky.clouds.runpod import RunPod
from sky.clouds.scp import SCP
from sky.clouds.vsphere import Vsphere

__all__ = [
    'IBM',
    'AWS',
    'Azure',
    'Cloud',
    'Cudo',
    'GCP',
    'Lambda',
    'SCP',
    'RunPod',
    'OCI',
    'Vsphere',
    'Kubernetes',
    'CloudImplementationFeatures',
    'Region',
    'Zone',
    'CLOUD_REGISTRY',
    'ProvisionerVersion',
    'StatusVersion',
    'Fluidstack',
    # Utility functions
    'cloud_in_list',
]



"""Kubernetes instance provisioning."""
import copy
import time
from typing import Any, Dict, List, Optional
import uuid

from sky import exceptions
from sky import sky_logging
from sky import skypilot_config
from sky import status_lib
from sky.adaptors import kubernetes
from sky.provision import common
from sky.provision.kubernetes import config as config_lib
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import common_utils
from sky.utils import kubernetes_enums
from sky.utils import ux_utils

POLL_INTERVAL = 2
_TIMEOUT_FOR_POD_TERMINATION = 60  # 1 minutes

logger = sky_logging.init_logger(__name__)
TAG_RAY_CLUSTER_NAME = 'ray-cluster-name'
TAG_SKYPILOT_CLUSTER_NAME = 'skypilot-cluster-name'
TAG_RAY_NODE_KIND = 'ray-node-type'  # legacy tag for backward compatibility
TAG_POD_INITIALIZED = 'skypilot-initialized'

POD_STATUSES = {
    'Pending', 'Running', 'Succeeded', 'Failed', 'Unknown', 'Terminating'
}


def to_label_selector(tags):
    label_selector = ''
    for k, v in tags.items():
        if label_selector != '':
            label_selector += ','
        label_selector += '{}={}'.format(k, v)
    return label_selector


def _get_namespace(provider_config: Dict[str, Any]) -> str:
    return provider_config.get(
        'namespace',
        kubernetes_utils.get_current_kube_config_context_namespace())


def _filter_pods(namespace: str, tag_filters: Dict[str, str],
                 status_filters: Optional[List[str]]) -> Dict[str, Any]:
    """Filters pods by tags and status."""
    non_included_pod_statuses = POD_STATUSES.copy()

    field_selector = ''
    if status_filters is not None:
        non_included_pod_statuses -= set(status_filters)
        field_selector = ','.join(
            [f'status.phase!={status}' for status in non_included_pod_statuses])

    label_selector = to_label_selector(tag_filters)
    pod_list = kubernetes.core_api().list_namespaced_pod(
        namespace, field_selector=field_selector, label_selector=label_selector)

    # Don't return pods marked for deletion,
    # i.e. pods with non-null metadata.DeletionTimestamp.
    pods = [
        pod for pod in pod_list.items if pod.metadata.deletion_timestamp is None
    ]
    return {pod.metadata.name: pod for pod in pods}


def _get_head_pod_name(pods: Dict[str, Any]) -> Optional[str]:
    head_pod_name = None
    for pod_name, pod in pods.items():
        if pod.metadata.labels[TAG_RAY_NODE_KIND] == 'head':
            head_pod_name = pod_name
            break
    return head_pod_name


def head_service_selector(cluster_name: str) -> Dict[str, str]:
    """Selector for Operator-configured head service."""
    return {'component': f'{cluster_name}-head'}


def _raise_pod_scheduling_errors(namespace, new_nodes):
    """Raise pod scheduling failure reason.

    When a pod fails to schedule in Kubernetes, the reasons for the failure
    are recorded as events. This function retrieves those events and raises
    descriptive errors for better debugging and user feedback.
    """
    for new_node in new_nodes:
        pod = kubernetes.core_api().read_namespaced_pod(new_node.metadata.name,
                                                        namespace)
        pod_status = pod.status.phase
        # When there are multiple pods involved while launching instance,
        # there may be a single pod causing issue while others are
        # successfully scheduled. In this case, we make sure to not surface
        # the error message from the pod that is already scheduled.
        if pod_status != 'Pending':
            continue
        pod_name = pod._metadata._name  # pylint: disable=protected-access
        events = kubernetes.core_api().list_namespaced_event(
            namespace,
            field_selector=(f'involvedObject.name={pod_name},'
                            'involvedObject.kind=Pod'))
        # Events created in the past hours are kept by
        # Kubernetes python client and we want to surface
        # the latest event message
        events_desc_by_time = sorted(
            events.items,
            key=lambda e: e.metadata.creation_timestamp,
            reverse=True)

        event_message = None
        for event in events_desc_by_time:
            if event.reason == 'FailedScheduling':
                event_message = event.message
                break
        timeout_err_msg = ('Timed out while waiting for nodes to start. '
                           'Cluster may be out of resources or '
                           'may be too slow to autoscale.')
        lack_resource_msg = (
            'Insufficient {resource} capacity on the cluster. '
            'Other SkyPilot tasks or pods may be using resources. '
            'Check resource usage by running `kubectl describe nodes`.')
        if event_message is not None:
            if pod_status == 'Pending':
                if 'Insufficient cpu' in event_message:
                    raise config_lib.KubernetesError(
                        lack_resource_msg.format(resource='CPU'))
                if 'Insufficient memory' in event_message:
                    raise config_lib.KubernetesError(
                        lack_resource_msg.format(resource='memory'))
                gpu_lf_keys = [
                    lf.get_label_key()
                    for lf in kubernetes_utils.LABEL_FORMATTER_REGISTRY
                ]
                if pod.spec.node_selector:
                    for label_key in pod.spec.node_selector.keys():
                        if label_key in gpu_lf_keys:
                            # TODO(romilb): We may have additional node
                            #  affinity selectors in the future - in that
                            #  case we will need to update this logic.
                            if (('Insufficient nvidia.com/gpu'
                                 in event_message) or
                                ('didn\'t match Pod\'s node affinity/selector'
                                 in event_message)):
                                msg = lack_resource_msg.format(resource='GPU')
                                raise config_lib.KubernetesError(
                                    f'{msg} Verify if '
                                    f'{pod.spec.node_selector[label_key]}'
                                    ' is available in the cluster.')
            raise config_lib.KubernetesError(f'{timeout_err_msg} '
                                             f'Pod status: {pod_status}'
                                             f'Details: \'{event_message}\' ')
    raise config_lib.KubernetesError(f'{timeout_err_msg}')


def _wait_for_pods_to_schedule(namespace, new_nodes, timeout: int):
    """Wait for all pods to be scheduled.

    Wait for all pods including jump pod to be scheduled, and if it
    exceeds the timeout, raise an exception. If pod's container
    is ContainerCreating, then we can assume that resources have been
    allocated and we can exit.
    """
    start_time = time.time()
    while time.time() - start_time < timeout:
        all_pods_scheduled = True
        for node in new_nodes:
            # Iterate over each pod to check their status
            pod = kubernetes.core_api().read_namespaced_pod(
                node.metadata.name, namespace)
            if pod.status.phase == 'Pending':
                # If container_statuses is None, then the pod hasn't
                # been scheduled yet.
                if pod.status.container_statuses is None:
                    all_pods_scheduled = False
                    break

        if all_pods_scheduled:
            return
        time.sleep(1)

    # Handle pod scheduling errors
    try:
        _raise_pod_scheduling_errors(namespace, new_nodes)
    except config_lib.KubernetesError:
        raise
    except Exception as e:
        raise config_lib.KubernetesError(
            'An error occurred while trying to fetch the reason '
            'for pod scheduling failure. '
            f'Error: {common_utils.format_exception(e)}') from None


def _wait_for_pods_to_run(namespace, new_nodes):
    """Wait for pods and their containers to be ready.

    Pods may be pulling images or may be in the process of container
    creation.
    """
    while True:
        all_pods_running = True
        # Iterate over each pod to check their status
        for node in new_nodes:
            pod = kubernetes.core_api().read_namespaced_pod(
                node.metadata.name, namespace)

            # Continue if pod and all the containers within the
            # pod are succesfully created and running.
            if pod.status.phase == 'Running' and all(
                    container.state.running
                    for container in pod.status.container_statuses):
                continue

            all_pods_running = False
            if pod.status.phase == 'Pending':
                # Iterate over each container in pod to check their status
                for container_status in pod.status.container_statuses:
                    # If the container wasn't in 'ContainerCreating'
                    # state, then we know pod wasn't scheduled or
                    # had some other error, such as image pull error.
                    # See list of possible reasons for waiting here:
                    # https://stackoverflow.com/a/57886025
                    waiting = container_status.state.waiting
                    if (waiting is not None and
                            waiting.reason != 'ContainerCreating'):
                        raise config_lib.KubernetesError(
                            'Failed to create container while launching '
                            'the node. Error details: '
                            f'{container_status.state.waiting.message}.')
            # Reaching this point means that one of the pods had an issue,
            # so break out of the loop
            break

        if all_pods_running:
            break
        time.sleep(1)


def _run_command_on_pods(node_name, node_namespace, command):
    cmd_output = kubernetes.stream()(
        kubernetes.core_api().connect_get_namespaced_pod_exec,
        node_name,
        node_namespace,
        command=command,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
        _request_timeout=kubernetes.API_TIMEOUT)
    return cmd_output


def _set_env_vars_in_pods(namespace: str, new_pods: List):
    """Setting environment variables in pods.

    Once all containers are ready, we can exec into them and set env vars.
    Kubernetes automatically populates containers with critical
    environment variables, such as those for discovering services running
    in the cluster and CUDA/nvidia environment variables. We need to
    make sure these env vars are available in every task and ssh session.
    This is needed for GPU support and service discovery.
    See https://github.com/skypilot-org/skypilot/issues/2287 for
    more details.

    To do so, we capture env vars from the pod's runtime and write them to
    /etc/profile.d/, making them available for all users in future
    shell sessions.
    """
    set_k8s_env_var_cmd = [
        '/bin/sh', '-c',
        ('prefix_cmd() '
         '{ if [ $(id -u) -ne 0 ]; then echo "sudo"; else echo ""; fi; } && '
         'printenv | awk -F "=" \'{print "export " $1 "=\\047" $2 "\\047"}\' > '
         '~/k8s_env_var.sh && '
         'mv ~/k8s_env_var.sh /etc/profile.d/k8s_env_var.sh || '
         '$(prefix_cmd) mv ~/k8s_env_var.sh /etc/profile.d/k8s_env_var.sh')
    ]

    for new_pod in new_pods:
        _run_command_on_pods(new_pod.metadata.name, namespace,
                             set_k8s_env_var_cmd)


def _check_user_privilege(namespace: str, new_nodes: List) -> None:
    # Checks if the default user has sufficient privilege to set up
    # the kubernetes instance pod.
    check_k8s_user_sudo_cmd = [
        '/bin/sh',
        '-c',
        (
            'if [ $(id -u) -eq 0 ]; then'
            # If user is root, create an alias for sudo used in skypilot setup
            '  echo \'alias sudo=""\' >> ~/.bashrc; '
            'else '
            '  if command -v sudo >/dev/null 2>&1; then '
            '    timeout 2 sudo -l >/dev/null 2>&1 || '
            f'    ( echo {exceptions.INSUFFICIENT_PRIVILEGES_CODE!r}; ); '
            '  else '
            f'    ( echo {exceptions.INSUFFICIENT_PRIVILEGES_CODE!r}; ); '
            '  fi; '
            'fi')
    ]

    for new_node in new_nodes:
        privilege_check = _run_command_on_pods(new_node.metadata.name,
                                               namespace,
                                               check_k8s_user_sudo_cmd)
        if privilege_check == str(exceptions.INSUFFICIENT_PRIVILEGES_CODE):
            raise config_lib.KubernetesError(
                'Insufficient system privileges detected. '
                'Ensure the default user has root access or '
                '"sudo" is installed and the user is added to the sudoers '
                'from the image.')


def _setup_ssh_in_pods(namespace: str, new_nodes: List) -> None:
    # Setting up ssh for the pod instance. This is already setup for
    # the jump pod so it does not need to be run for it.
    set_k8s_ssh_cmd = [
        '/bin/sh',
        '-c',
        (
            'prefix_cmd() '
            '{ if [ $(id -u) -ne 0 ]; then echo "sudo"; else echo ""; fi; }; '
            'export DEBIAN_FRONTEND=noninteractive;'
            '$(prefix_cmd) apt-get update;'
            '$(prefix_cmd) apt install openssh-server rsync -y; '
            '$(prefix_cmd) mkdir -p /var/run/sshd; '
            '$(prefix_cmd) '
            'sed -i "s/PermitRootLogin prohibit-password/PermitRootLogin yes/" '
            '/etc/ssh/sshd_config; '
            '$(prefix_cmd) sed '
            '"s@session\\s*required\\s*pam_loginuid.so@session optional '
            'pam_loginuid.so@g" -i /etc/pam.d/sshd; '
            'cd /etc/ssh/ && $(prefix_cmd) ssh-keygen -A; '
            '$(prefix_cmd) mkdir -p ~/.ssh; '
            '$(prefix_cmd) chown -R $(whoami) ~/.ssh;'
            '$(prefix_cmd) chmod 700 ~/.ssh; '
            '$(prefix_cmd) chmod 644 ~/.ssh/authorized_keys; '
            '$(prefix_cmd) cat /etc/secret-volume/ssh-publickey* > '
            '~/.ssh/authorized_keys; '
            '$(prefix_cmd) service ssh restart; '
            # Eliminate the error
            # `mesg: ttyname failed: inappropriate ioctl for device`.
            # See https://www.educative.io/answers/error-mesg-ttyname-failed-inappropriate-ioctl-for-device  # pylint: disable=line-too-long
            '$(prefix_cmd) sed -i "s/mesg n/tty -s \\&\\& mesg n/" ~/.profile;')
    ]
    # TODO(romilb): We need logging and surface errors here.
    for new_node in new_nodes:
        _run_command_on_pods(new_node.metadata.name, namespace, set_k8s_ssh_cmd)


def _label_pod(namespace: str, pod_name: str, label: Dict[str, str]) -> None:
    """Label a pod."""
    kubernetes.core_api().patch_namespaced_pod(
        pod_name,
        namespace, {'metadata': {
            'labels': label
        }},
        _request_timeout=kubernetes.API_TIMEOUT)


def _create_pods(region: str, cluster_name_on_cloud: str,
                 config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Create pods based on the config."""
    provider_config = config.provider_config
    namespace = _get_namespace(provider_config)
    pod_spec = copy.deepcopy(config.node_config)
    tags = {
        TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud,
    }
    pod_spec['metadata']['namespace'] = namespace
    if 'labels' in pod_spec['metadata']:
        pod_spec['metadata']['labels'].update(tags)
    else:
        pod_spec['metadata']['labels'] = tags
    pod_spec['metadata']['labels'].update(
        {TAG_SKYPILOT_CLUSTER_NAME: cluster_name_on_cloud})

    terminating_pods = _filter_pods(namespace, tags, ['Terminating'])
    start_time = time.time()
    while (len(terminating_pods) > 0 and
           time.time() - start_time < _TIMEOUT_FOR_POD_TERMINATION):
        logger.debug(f'run_instances: Found {len(terminating_pods)} '
                     'terminating pods. Waiting them to finish: '
                     f'{list(terminating_pods.keys())}')
        time.sleep(POLL_INTERVAL)
        terminating_pods = _filter_pods(namespace, tags, ['Terminating'])

    if len(terminating_pods) > 0:
        # If there are still terminating pods, we force delete them.
        logger.debug(f'run_instances: Found {len(terminating_pods)} '
                     'terminating pods still in terminating state after '
                     f'timeout {_TIMEOUT_FOR_POD_TERMINATION}s. '
                     'Force deleting them.')
        for pod_name in terminating_pods.keys():
            # grace_period_seconds=0 means force delete the pod.
            # https://github.com/kubernetes-client/python/issues/508#issuecomment-1695759777
            kubernetes.core_api().delete_namespaced_pod(
                pod_name,
                namespace,
                _request_timeout=config_lib.DELETION_TIMEOUT,
                grace_period_seconds=0)

    running_pods = _filter_pods(namespace, tags, ['Pending', 'Running'])
    head_pod_name = _get_head_pod_name(running_pods)
    logger.debug(f'Found {len(running_pods)} existing pods: '
                 f'{list(running_pods.keys())}')

    to_start_count = config.count - len(running_pods)
    if to_start_count < 0:
        raise RuntimeError(
            'The number of running+pending pods '
            f'({config.count - to_start_count}) in cluster '
            f'"{cluster_name_on_cloud}" is greater than the number '
            f'requested by the user ({config.count}). '
            'This is likely a resource leak. '
            'Use "sky down" to terminate the cluster.')

    # Add nvidia runtime class if it exists
    nvidia_runtime_exists = False
    try:
        nvidia_runtime_exists = kubernetes_utils.check_nvidia_runtime_class()
    except kubernetes.get_kubernetes().client.ApiException as e:
        logger.warning('run_instances: Error occurred while checking for '
                       f'nvidia RuntimeClass - '
                       f'{common_utils.format_exception(e)}'
                       'Continuing without using nvidia RuntimeClass.\n'
                       'If you are on a K3s cluster, manually '
                       'override runtimeClassName in ~/.sky/config.yaml. '
                       'For more details, refer to https://skypilot.readthedocs.io/en/latest/reference/config.html')  # pylint: disable=line-too-long

    if nvidia_runtime_exists:
        pod_spec['spec']['runtimeClassName'] = 'nvidia'

    created_pods = {}
    logger.debug(f'run_instances: calling create_namespaced_pod '
                 f'(count={to_start_count}).')
    for _ in range(to_start_count):
        if head_pod_name is None:
            pod_spec['metadata']['labels'][TAG_RAY_NODE_KIND] = 'head'
            head_selector = head_service_selector(cluster_name_on_cloud)
            pod_spec['metadata']['labels'].update(head_selector)
            pod_spec['metadata']['name'] = f'{cluster_name_on_cloud}-head'
        else:
            pod_spec['metadata']['labels'][TAG_RAY_NODE_KIND] = 'worker'
            pod_uuid = str(uuid.uuid4())[:4]
            pod_name = f'{cluster_name_on_cloud}-{pod_uuid}'
            pod_spec['metadata']['name'] = f'{pod_name}-worker'
            # For multi-node support, we put a soft-constraint to schedule
            # worker pods on different nodes than the head pod.
            # This is not set as a hard constraint because if different nodes
            # are not available, we still want to be able to schedule worker
            # pods on larger nodes which may be able to fit multiple SkyPilot
            # "nodes".
            pod_spec['spec']['affinity'] = {
                'podAntiAffinity': {
                    'requiredDuringSchedulingIgnoredDuringExecution': [{
                        'labelSelector': {
                            'matchExpressions': [{
                                'key': TAG_SKYPILOT_CLUSTER_NAME,
                                'operator': 'In',
                                'values': [cluster_name_on_cloud]
                            }]
                        },
                        'topologyKey': 'kubernetes.io/hostname'
                    }]
                }
            }
        pod = kubernetes.core_api().create_namespaced_pod(namespace, pod_spec)
        created_pods[pod.metadata.name] = pod
        if head_pod_name is None:
            head_pod_name = pod.metadata.name

    # Adding the jump pod to the new_nodes list as well so it can be
    # checked if it's scheduled and running along with other pods.
    ssh_jump_pod_name = pod_spec['metadata']['labels']['skypilot-ssh-jump']
    jump_pod = kubernetes.core_api().read_namespaced_pod(
        ssh_jump_pod_name, namespace)
    wait_pods_dict = _filter_pods(namespace, tags, ['Pending'])
    wait_pods = list(wait_pods_dict.values())
    wait_pods.append(jump_pod)
    logger.debug('run_instances: waiting for pods to schedule and run: '
                 f'{list(wait_pods_dict.keys())}')

    # Wait until the pods are scheduled and surface cause for error
    # if there is one
    _wait_for_pods_to_schedule(namespace, wait_pods, provider_config['timeout'])
    # Wait until the pods and their containers are up and running, and
    # fail early if there is an error
    _wait_for_pods_to_run(namespace, wait_pods)
    logger.debug(f'run_instances: all pods are scheduled and running: '
                 f'{list(wait_pods_dict.keys())}')

    running_pods = _filter_pods(namespace, tags, ['Running'])
    initialized_pods = _filter_pods(namespace, {
        TAG_POD_INITIALIZED: 'true',
        **tags
    }, ['Running'])
    uninitialized_pods = {
        pod_name: pod
        for pod_name, pod in running_pods.items()
        if pod_name not in initialized_pods
    }
    if len(uninitialized_pods) > 0:
        logger.debug(f'run_instances: Initializing {len(uninitialized_pods)} '
                     f'pods: {list(uninitialized_pods.keys())}')
        uninitialized_pods_list = list(uninitialized_pods.values())
        _check_user_privilege(namespace, uninitialized_pods_list)
        _setup_ssh_in_pods(namespace, uninitialized_pods_list)
        _set_env_vars_in_pods(namespace, uninitialized_pods_list)
        for pod in uninitialized_pods.values():
            _label_pod(namespace,
                       pod.metadata.name,
                       label={
                           TAG_POD_INITIALIZED: 'true',
                           **pod.metadata.labels
                       })

    assert head_pod_name is not None, 'head_instance_id should not be None'
    return common.ProvisionRecord(
        provider_name='kubernetes',
        region=region,
        zone=None,
        cluster_name=cluster_name_on_cloud,
        head_instance_id=head_pod_name,
        resumed_instance_ids=[],
        created_instance_ids=list(created_pods.keys()),
    )


def run_instances(region: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Runs instances for the given cluster."""
    try:
        return _create_pods(region, cluster_name_on_cloud, config)
    except (kubernetes.api_exception(), config_lib.KubernetesError) as e:
        logger.warning(f'run_instances: Error occurred when creating pods: {e}')
        raise


def wait_instances(region: str, cluster_name_on_cloud: str,
                   state: Optional[status_lib.ClusterStatus]) -> None:
    del region, cluster_name_on_cloud, state


def stop_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    raise NotImplementedError()


def _terminate_node(namespace: str, pod_name: str) -> None:
    """Terminate a pod."""
    logger.debug('terminate_instances: calling delete_namespaced_pod')
    try:
        kubernetes_utils.clean_zombie_ssh_jump_pod(namespace, pod_name)
    except Exception as e:  # pylint: disable=broad-except
        logger.warning('terminate_instances: Error occurred when analyzing '
                       f'SSH Jump pod: {e}')
    try:
        kubernetes.core_api().delete_namespaced_service(
            pod_name, namespace, _request_timeout=config_lib.DELETION_TIMEOUT)
        kubernetes.core_api().delete_namespaced_service(
            f'{pod_name}-ssh',
            namespace,
            _request_timeout=config_lib.DELETION_TIMEOUT)
    except kubernetes.api_exception():
        pass
    # Note - delete pod after all other resources are deleted.
    # This is to ensure there are no leftover resources if this down is run
    # from within the pod, e.g., for autodown.
    try:
        kubernetes.core_api().delete_namespaced_pod(
            pod_name, namespace, _request_timeout=config_lib.DELETION_TIMEOUT)
    except kubernetes.api_exception() as e:
        if e.status == 404:
            logger.warning('terminate_instances: Tried to delete pod '
                           f'{pod_name}, but the pod was not found (404).')
        else:
            raise


def terminate_instances(
    cluster_name_on_cloud: str,
    provider_config: Dict[str, Any],
    worker_only: bool = False,
) -> None:
    """See sky/provision/__init__.py"""
    namespace = _get_namespace(provider_config)
    tag_filters = {
        TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud,
    }
    pods = _filter_pods(namespace, tag_filters, None)

    def _is_head(pod) -> bool:
        return pod.metadata.labels[TAG_RAY_NODE_KIND] == 'head'

    for pod_name, pod in pods.items():
        logger.debug(f'Terminating instance {pod_name}: {pod}')
        if _is_head(pod) and worker_only:
            continue
        _terminate_node(namespace, pod_name)


def get_cluster_info(
        region: str,
        cluster_name_on_cloud: str,
        provider_config: Optional[Dict[str, Any]] = None) -> common.ClusterInfo:
    del region  # unused
    assert provider_config is not None
    namespace = _get_namespace(provider_config)
    tag_filters = {
        TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud,
    }

    running_pods = _filter_pods(namespace, tag_filters, ['Running'])
    pods: Dict[str, List[common.InstanceInfo]] = {}
    head_pod_name = None

    port_forward_mode = kubernetes_enums.KubernetesNetworkingMode.PORTFORWARD
    network_mode_str = skypilot_config.get_nested(('kubernetes', 'networking'),
                                                  port_forward_mode.value)
    network_mode = kubernetes_enums.KubernetesNetworkingMode.from_str(
        network_mode_str)
    external_ip = kubernetes_utils.get_external_ip(network_mode)
    port = 22
    if not provider_config.get('use_internal_ips', False):
        port = kubernetes_utils.get_head_ssh_port(cluster_name_on_cloud,
                                                  namespace)

    head_pod_name = None
    cpu_request = None
    for pod_name, pod in running_pods.items():
        internal_ip = pod.status.pod_ip
        pods[pod_name] = [
            common.InstanceInfo(
                instance_id=pod_name,
                internal_ip=internal_ip,
                external_ip=(None if network_mode == port_forward_mode else
                             external_ip),
                ssh_port=port,
                tags=pod.metadata.labels,
            )
        ]
        if pod.metadata.labels[TAG_RAY_NODE_KIND] == 'head':
            head_pod_name = pod_name
            head_spec = pod.spec
            assert head_spec is not None, pod
            cpu_request = head_spec.containers[0].resources.requests['cpu']

    assert cpu_request is not None, 'cpu_request should not be None'

    ssh_user = 'sky'
    get_k8s_ssh_user_cmd = ['/bin/sh', '-c', ('echo $(whoami)')]
    assert head_pod_name is not None
    ssh_user = _run_command_on_pods(head_pod_name, namespace,
                                    get_k8s_ssh_user_cmd)
    ssh_user = ssh_user.strip()
    logger.debug(
        f'Using ssh user {ssh_user} for cluster {cluster_name_on_cloud}')

    return common.ClusterInfo(
        instances=pods,
        head_instance_id=head_pod_name,
        ssh_user=ssh_user,
        # We manually set object-store-memory=500000000 to avoid ray from
        # allocating a very large object store in each pod that may cause
        # problems for other pods.
        custom_ray_options={
            'object-store-memory': 500000000,
            'num-cpus': cpu_request,
        })


def query_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True
) -> Dict[str, Optional[status_lib.ClusterStatus]]:
    del provider_config  # unused
    status_map = {
        'Pending': status_lib.ClusterStatus.INIT,
        'Running': status_lib.ClusterStatus.UP,
        'Failed': None,
        'Unknown': None,
        'Succeeded': None,
        'Terminating': None,
    }

    namespace = kubernetes_utils.get_current_kube_config_context_namespace()

    # Get all the pods with the label skypilot-cluster: <cluster_name>
    try:
        pods = kubernetes.core_api().list_namespaced_pod(
            namespace,
            label_selector=f'skypilot-cluster={cluster_name_on_cloud}',
            _request_timeout=kubernetes.API_TIMEOUT).items
    except kubernetes.max_retry_error():
        with ux_utils.print_exception_no_traceback():
            ctx = kubernetes_utils.get_current_kube_config_context_name()
            raise exceptions.ClusterStatusFetchingError(
                f'Failed to query cluster {cluster_name_on_cloud!r} status. '
                'Network error - check if the Kubernetes cluster in '
                f'context {ctx} is up and accessible.') from None
    except Exception as e:  # pylint: disable=broad-except
        with ux_utils.print_exception_no_traceback():
            raise exceptions.ClusterStatusFetchingError(
                f'Failed to query Kubernetes cluster {cluster_name_on_cloud!r} '
                f'status: {common_utils.format_exception(e)}')

    # Check if the pods are running or pending
    cluster_status = {}
    for pod in pods:
        pod_status = status_map[pod.status.phase]
        if non_terminated_only and pod_status is None:
            continue
        cluster_status[pod.metadata.name] = pod_status
    return cluster_status



"""Constants used for Managed Spot."""

SPOT_CONTROLLER_TEMPLATE = 'spot-controller.yaml.j2'
SPOT_CONTROLLER_YAML_PREFIX = '~/.sky/spot_controller'

SPOT_TASK_YAML_PREFIX = '~/.sky/spot_tasks'

# Resources as a dict for the spot controller.
# Use default CPU instance type for spot controller with >= 24GB, i.e.
# m6i.2xlarge (8vCPUs, 32 GB) for AWS, Standard_D8s_v4 (8vCPUs, 32 GB)
# for Azure, and n1-standard-8 (8 vCPUs, 32 GB) for GCP, etc.
# Based on profiling, memory should be at least 3x (in GB) as num vCPUs to avoid
# OOM (each vCPU can have 4 spot controller processes as we set the CPU
# requirement to 0.25, and 3 GB is barely enough for 4 spot processes).
# We use 50 GB disk size to reduce the cost.
CONTROLLER_RESOURCES = {'cpus': '8+', 'memory': '3x', 'disk_size': 50}

# Max length of the cluster name for GCP is 35, the user hash to be attached is
# 4+1 chars, and we assume the maximum length of the job id is 4+1, so the max
# length of the cluster name prefix is 25 to avoid the cluster name being too
# long and truncated twice during the cluster creation.
SPOT_CLUSTER_NAME_PREFIX_LENGTH = 25



"""ID Generator
"""

import random
import string
import uuid


def generate_random_uuid():
    return str(uuid.uuid4())


def rand(value):
    return value + generate_random_string(5)


def generate_random_string(length):
    return ''.join(random.choice(string.ascii_uppercase) for _ in range(length))



"""Azure instance provisioning."""
import logging
from typing import Any, Callable, Dict, List, Optional

from sky import sky_logging
from sky.adaptors import azure
from sky.provision import common
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

# Suppress noisy logs from Azure SDK. Reference:
# https://github.com/Azure/azure-sdk-for-python/issues/9422
azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.WARNING)

# Tag uniquely identifying all nodes of a cluster
TAG_RAY_CLUSTER_NAME = 'ray-cluster-name'
TAG_RAY_NODE_KIND = 'ray-node-type'


def get_azure_sdk_function(client: Any, function_name: str) -> Callable:
    """Retrieve a callable function from Azure SDK client object.

    Newer versions of the various client SDKs renamed function names to
    have a begin_ prefix. This function supports both the old and new
    versions of the SDK by first trying the old name and falling back to
    the prefixed new name.
    """
    func = getattr(client, function_name,
                   getattr(client, f'begin_{function_name}', None))
    if func is None:
        raise AttributeError(
            '"{obj}" object has no {func} or begin_{func} attribute'.format(
                obj={client.__name__}, func=function_name))
    return func


def open_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """See sky/provision/__init__.py"""
    assert provider_config is not None, cluster_name_on_cloud
    subscription_id = provider_config['subscription_id']
    resource_group = provider_config['resource_group']
    network_client = azure.get_client('network', subscription_id)
    # The NSG should have been created by the cluster provisioning.
    update_network_security_groups = get_azure_sdk_function(
        client=network_client.network_security_groups,
        function_name='create_or_update')
    list_network_security_groups = get_azure_sdk_function(
        client=network_client.network_security_groups, function_name='list')
    for nsg in list_network_security_groups(resource_group):
        try:
            # Azure NSG rules have a priority field that determines the order
            # in which they are applied. The priority must be unique across
            # all inbound rules in one NSG.
            priority = max(rule.priority
                           for rule in nsg.security_rules
                           if rule.direction == 'Inbound') + 1
            nsg.security_rules.append(
                azure.create_security_rule(
                    name=f'sky-ports-{cluster_name_on_cloud}-{priority}',
                    priority=priority,
                    protocol='Tcp',
                    access='Allow',
                    direction='Inbound',
                    source_address_prefix='*',
                    source_port_range='*',
                    destination_address_prefix='*',
                    destination_port_ranges=ports,
                ))
            poller = update_network_security_groups(resource_group, nsg.name,
                                                    nsg)
            poller.wait()
            if poller.status() != 'Succeeded':
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(f'Failed to open ports {ports} in NSG '
                                     f'{nsg.name}: {poller.status()}')
        except azure.http_error_exception() as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to open ports {ports} in NSG {nsg.name}.') from e


def cleanup_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """See sky/provision/__init__.py"""
    # Azure will automatically cleanup network security groups when cleanup
    # resource group. So we don't need to do anything here.
    del cluster_name_on_cloud, ports, provider_config  # Unused.


def query_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> Dict[int, List[common.Endpoint]]:
    """See sky/provision/__init__.py"""
    return common.query_ports_passthrough(cluster_name_on_cloud, ports,
                                          provider_config)



"""Configuration for service catalog."""

import contextlib
import functools
import threading

_thread_local_config = threading.local()


@contextlib.contextmanager
def _set_use_default_catalog_if_failed(value: bool):
    old_value = get_use_default_catalog_if_failed()
    _thread_local_config.use_default_catalog = value
    try:
        yield
    finally:
        _thread_local_config.use_default_catalog = old_value


def get_use_default_catalog_if_failed() -> bool:
    """Whether to use default catalog if failed to fetch account-specific one.

    Whether the caller requires the catalog to be narrowed down to the account-
    specific catalog (e.g., removing regions not enabled for the current account
    or use zone name assigned to the AWS account).

    When set to True, the caller allows to use the default service catalog,
    which may have inaccurate information (e.g., AWS's zone names are account-
    specific), but it is ok for the read-only operators, such as `show-gpus` or
    `sky status`.
    """
    if not hasattr(_thread_local_config, 'use_default_catalog'):
        # Should not set it globally, as the global assignment
        # will be executed only once if the module is imported
        # in the main thread, and will not be executed in other
        # threads.
        _thread_local_config.use_default_catalog = False
    return _thread_local_config.use_default_catalog


def fallback_to_default_catalog(func):
    """Decorator: allow failure for fetching account-specific catalog.

    The account-specific catalog requires the credentials of the
    cloud to be set.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with _set_use_default_catalog_if_failed(True):
            return func(*args, **kwargs)

    return wrapper
